@@ -1,3 +1,4 @@
+import sys
 from collections import defaultdict
 from typing import Optional, List, Set, Any, Union, Dict, Tuple
 from dataclasses import dataclass, field
@@ -5,6 +6,7 @@ import logging
 
 import click
 from funowl.converters.functional_converter import to_python
+from funowl.writers.FunctionalWriter import FunctionalWriter
 from jinja2 import Template
 from linkml.generators.pythongen import PythonGenerator
 from linkml_runtime import SchemaView
@@ -14,7 +16,7 @@ from linkml_runtime.utils.eval_utils import eval_expr
 from linkml_runtime.index.object_index import ObjectIndex
 from linkml_runtime.utils.inference_utils import infer_all_slot_values, infer_slot_value, Config
 
-from rdflib import URIRef
+from rdflib import URIRef, Graph
 
 from linkml_runtime.linkml_model.meta import ClassDefinition, SchemaDefinition, SlotDefinition, Definition, \
     ClassDefinitionName
@@ -30,7 +32,8 @@ from funowl import OntologyDocument, Ontology, IRI, ObjectSomeValuesFrom, \
     ObjectProperty, Declaration, InverseObjectProperties, ObjectPropertyDomain, ObjectPropertyRange, \
     SubObjectPropertyOf, TransitiveObjectProperty, SymmetricObjectProperty, AsymmetricObjectProperty, \
     ReflexiveObjectProperty, IrreflexiveObjectProperty, Annotation, ObjectMinCardinality, ObjectHasValue, \
-    NamedIndividual, DataSomeValuesFrom, DataHasValue, DataAllValuesFrom
+    NamedIndividual, DataSomeValuesFrom, DataHasValue, DataAllValuesFrom, AnnotationProperty, DataProperty, Datatype, \
+    DisjointClasses
 
 from linkml_runtime.dumpers.dumper_root import Dumper
 from linkml_runtime.utils.yamlutils import YAMLRoot
@@ -134,7 +137,7 @@ class OWLDumper(Dumper):
     object_index: ObjectIndex = None
     """Index of LinkML objects."""
 
-    infer_missing_values: bool = False
+    autofill: bool = False
     """If True, infer missing values in data."""
 
     def to_ontology_document(self, element: Union[YAMLRoot, List[YAMLRoot]], schema: Union[SchemaDefinition, str],
@@ -166,7 +169,7 @@ class OWLDumper(Dumper):
             doc.prefixDeclarations.append(Prefix(pfx.prefix_prefix, pfx.prefix_reference))
         return doc
 
-    def dumps(self, element: YAMLRoot, schema: SchemaDefinition = None, schemaview: SchemaView = None, iri=None) -> str:
+    def dumps(self, element: YAMLRoot, schema: SchemaDefinition = None, schemaview: SchemaView = None, iri=None, output_type=None) -> str:
         """
         Dump a linkml instance tree to a function syntax OWL ontology string
 
@@ -174,15 +177,20 @@ class OWLDumper(Dumper):
         :param schema:
         :param schemaview:
         :param iri:
+        :param output_type:
         :return:
         """
         if schemaview:
             schema = schemaview.schema
         doc = self.to_ontology_document(element, schema, iri=iri)
-        return str(doc)
+        if output_type == "ttl":
+            g = Graph()
+            doc.to_rdf(g)
+            return g.serialize(format="ttl")
+        else:
+            return str(doc)
 
-    def transform(self, element: YAMLRoot, schema: SchemaDefinition, is_element_an_object=True,
-                  is_element_an_owl_class=False) -> Any:
+    def transform(self, element: YAMLRoot, schema: SchemaDefinition, is_element_an_object=True) -> Any:
         """
         Recursively transform a LinkML element
 
@@ -195,6 +203,7 @@ class OWLDumper(Dumper):
         :param is_element_an_owl_class:
         :return: IRI or node of transformation of element
         """
+        logging.debug(f"transform: {element}")
         if element is None:
             return None
         try:
@@ -218,7 +227,7 @@ class OWLDumper(Dumper):
         o = self.ontology
         python_type = type(element)
         linkml_class_name = python_type.class_name
-        if self.infer_missing_values:
+        if self.autofill:
             if not self.object_index:
                 raise ValueError("object_index must be set if infer_missing_values is True")
             proxy = self.object_index.bless(element)
@@ -232,12 +241,11 @@ class OWLDumper(Dumper):
         for fstr in self._get_inferred_class_annotations(c, 'owl.fstring'):
             self.add_axioms_from_fstring(fstr, element)
         for tmpl_str in self._get_inferred_class_annotations(c, 'owl.template'):
-            self.add_axioms_from_template(tmpl_str, element)
+            self.add_axioms_from_template(tmpl_str, element, schema=schema)
         cls_interps = self._get_class_interpretations(c)
-        if Class.__name__ in cls_interps:
-            is_element_an_owl_class = True
         subj = None
         eai = EntityAxiomIndex()
+        unprocessed_parents = []
         # set subj = IRI for element
         for k, v in vars(element).items():
             slot: SlotDefinition
@@ -246,26 +254,34 @@ class OWLDumper(Dumper):
                 raise ValueError(f'Lookup slot in {c.name} failed for {k} // element={element}')
             if slot.identifier:
                 subj = URIRef(self._get_IRI_str(v))
-        if subj is None:
+                logging.debug(f"set subj from {slot.name}={v} asURI: {subj}")
+        if AnonymousIndividual.__name__ in cls_interps or (subj is None and "Individual" in cls_interps):
             subj = AnonymousIndividual()
-        if ObjectProperty.__name__ in cls_interps:
-            # TODO: make this generic for all declarations
-            decl = Declaration(ObjectProperty(subj))
-            o.axioms.append(decl)
-        if NamedIndividual.__name__ in cls_interps:
-            # TODO: make this generic for all declarations
-            decl = Declaration(NamedIndividual(subj))
-            o.axioms.append(decl)
-        if Class.__name__ in cls_interps and not isinstance(subj, AnonymousIndividual):
-            # TODO: make this generic for all declarations
-            decl = Declaration(Class(subj))
-            o.axioms.append(decl)
+            logging.debug(f"No identifier slot; Creating anon individual for element = {subj}")
+        else:
+            for obj_type in [ObjectProperty, DataProperty, AnnotationProperty, NamedIndividual, Class, Datatype]:
+                if obj_type.__name__ in cls_interps:
+                    decl = Declaration(obj_type(subj))
+                    logging.debug(f"Inferred {decl} based on {obj_type.__name__} in {cls_interps}")
+                    o.axioms.append(decl)
+        logging.info(f"Subject={subj}")
+        expression_termset = {"IntersectionOf", "UnionOf", "ComplementOf", "OneOf", "SomeValuesFrom", "AllValuesFrom"}
+        is_returns_expression = len(expression_termset.intersection(cls_interps)) > 0
         # iterate through all slot-value assignments for element;
         # generate axioms or add axioms to EntityAxiomIndex for each
         for k, v in vars(element).items():
             slot: SlotDefinition
             # TODO: unify slot/schema_level_slot
             slot = self._lookup_slot(c, k)
+            if slot is None:
+                logging.error(f'No slot for {k}')
+                continue
+            if slot.identifier:
+                # the role of the identifier slot is to determine the IRI for the element;
+                # it generates no axioms of its own
+                continue
+            if 'owl.ignore' in slot.annotations:
+                continue
             schema_level_slot = self._get_schema_level_slot(slot)
             # lookup OWL settings on each slot
             owl_templates = self._get_inferred_slot_annotations(slot, 'owl.template', linkml_class_name)
@@ -285,26 +301,16 @@ class OWLDumper(Dumper):
                             ann_val = self.transform(ann_val, schema, is_element_an_object=ann_slot.range in self.schema.classes)
                             ann_slot_iri = self._get_IRI_str(ann_slot.slot_uri)
                             axiom_annotations.append(Annotation(ann_slot_iri, ann_val))
+            # templates
             for tmpl in owl_templates:
-                self.add_axioms_from_template(tmpl, element)
-            if slot is None:
-                logging.error(f'No slot for {k}')
-                continue
-            if slot.identifier:
-                # the role of the identifier slot is to determine the IRI for the element;
-                # it generates no axioms of its own
-                continue
+                self.add_axioms_from_template(tmpl, element, schema=schema)
             if schema_level_slot.slot_uri is not None:
                 slot_uri = self._get_IRI_str(schema_level_slot.slot_uri)
             else:
                 slot_uri = self._get_IRI_str(self.schemaview.get_uri(slot.name))
-            logging.debug(f'in {subj} {k} = {v} (URI={schema_level_slot.slot_uri}) // slot = {slot.name}')
-            interps = self._get_slot_interpretations(slot, linkml_class_name)
-            logging.debug(f'INTERPS={interps}')
-            # TODO: make this more generic
-            is_disjunction = 'UnionOf' in interps
-            is_conjunction = 'IntersectionOf' in interps
-            is_annotation = 'AnnotationProperty' in interps or 'Annotation' in interps
+            logging.debug(f'SlotVal {subj}.{k} = {v} (URI={schema_level_slot.slot_uri}) // slot = {slot.name}')
+            slot_interps = self._get_slot_interpretations(slot, linkml_class_name)
+            logging.debug(f'OWL interpretations for {k}={slot_interps}')
             is_object_ref = slot.range in self.schema.classes
             # normalize input_vals to a list, then recursively transform
             if isinstance(v, list):
@@ -317,6 +323,10 @@ class OWLDumper(Dumper):
             logging.debug(f'TR Vals={tr_vals}')
             parents = []  ## expressions that are the referents of the axioms to be generated
             is_class_logical_axiom = False
+            # TODO: make this more generic
+            is_disjunction = 'UnionOf' in slot_interps
+            is_conjunction = 'IntersectionOf' in slot_interps
+            is_annotation = 'AnnotationProperty' in slot_interps or 'Annotation' in slot_interps
             closure_axiom_parents = []
             for tr_val in tr_vals:
                 logging.debug(f'  TR_VAL = {tr_val}')
@@ -326,7 +336,7 @@ class OWLDumper(Dumper):
                     for owl_fstring in owl_fstrings:
                         self.add_axioms_from_fstring(owl_fstring, element, tr_val)
                     continue
-                if slot.range in self.schema.classes:
+                if is_object_ref:
                     if isinstance(tr_val, str):
                         tr_val = self._get_IRI_str(tr_val)
                 parent = tr_val
@@ -348,32 +358,16 @@ class OWLDumper(Dumper):
                                 raise ValueError(f'Cannot interpret {owl_uri}')
                 # transform parents if an expression type is specified
                 # TODO: use a mapping rather than repetitive code
-                if ObjectSomeValuesFrom.__name__ in interps:
-                    parent = ObjectSomeValuesFrom(slot_uri, tr_val)
-                    is_class_logical_axiom = True
-                elif DataSomeValuesFrom.__name__ in interps:
-                    parent = DataSomeValuesFrom(slot_uri, tr_val)
-                    is_class_logical_axiom = True
-                elif ObjectHasValue.__name__ in interps:
-                    parent = ObjectHasValue(slot_uri, tr_val)
-                    is_class_logical_axiom = True
-                elif DataHasValue.__name__ in interps:
-                    parent = DataHasValue(slot_uri, tr_val)
-                    is_class_logical_axiom = True
-                elif ObjectAllValuesFrom.__name__ in interps:
-                    parent = ObjectAllValuesFrom(slot_uri, tr_val)
-                    is_class_logical_axiom = True
-                elif DataAllValuesFrom.__name__ in interps:
-                    parent = DataAllValuesFrom(slot_uri, tr_val)
-                    is_class_logical_axiom = True
-                elif ObjectMinCardinality.__name__ in interps:
-                    raise NotImplementedError(f'TODO: QCRs')
-                elif SubClassOf.__name__ in interps:
-                    is_class_logical_axiom = True
-                elif EquivalentClasses.__name__ in interps:
-                    is_class_logical_axiom = True
+                simple_expression_types = [ObjectSomeValuesFrom, DataSomeValuesFrom, ObjectHasValue, DataHasValue, ObjectAllValuesFrom, DataAllValuesFrom]
+                for expression_type in simple_expression_types:
+                    if expression_type.__name__ in slot_interps:
+                        parent = expression_type(slot_uri, tr_val)
+                        is_class_logical_axiom = True
+                for logical_axiom_type in [SubClassOf, EquivalentClasses, DisjointClasses]:
+                    if logical_axiom_type.__name__ in slot_interps:
+                        is_class_logical_axiom = True
                 parents.append(parent)
-                if 'Closed' in interps:
+                if 'Closed' in slot_interps:
                     closure_axiom_parents.append(ObjectAllValuesFrom(slot_uri, tr_val))
             if closure_axiom_parents:
                 if len(closure_axiom_parents) == 1:
@@ -384,27 +378,14 @@ class OWLDumper(Dumper):
             #    eai.add_operands((0, SubClassOf.__name__, ObjectUnionOf.__name__), parents, axiom_annotations)
             axiom_type = None
             # TODO: make this more generic / less repetitive
-            if SubClassOf.__name__ in interps:
-                axiom_type = SubClassOf
-            elif SubObjectPropertyOf.__name__ in interps:
-                axiom_type = SubObjectPropertyOf
-            elif ClassAssertion.__name__ in interps:
-                axiom_type = ClassAssertion
-            elif ObjectPropertyAssertion.__name__ in interps:
-                axiom_type = ObjectPropertyAssertion
-            elif EquivalentClasses.__name__ in interps:
-                axiom_type = EquivalentClasses
-            elif InverseObjectProperties.__name__ in interps:
-                axiom_type = InverseObjectProperties
-            elif ObjectPropertyDomain.__name__ in interps:
-                axiom_type = ObjectPropertyDomain
-            elif ObjectPropertyRange.__name__ in interps:
-                axiom_type = ObjectPropertyRange
-            elif AnnotationAssertion.__name__ in interps:
-                axiom_type = AnnotationAssertion
+            axiom_types = [SubClassOf, SubObjectPropertyOf, ClassAssertion, ObjectPropertyAssertion, EquivalentClasses, InverseObjectProperties,ObjectPropertyDomain, ObjectPropertyRange, AnnotationAssertion]
+            for candidate_axiom_type in axiom_types:
+                if candidate_axiom_type.__name__ in slot_interps:
+                    axiom_type = candidate_axiom_type
+            # fill in default axiom type
             if axiom_type is None:
                 # default: SubClassOf R some V for logical; otherwise annotation
-                if is_class_logical_axiom:
+                if is_class_logical_axiom and not is_returns_expression:
                     if is_annotation:
                         raise ValueError(f'{slot.name} cannot be both logical and an annotation')
                     axiom_type = SubClassOf
@@ -413,73 +394,110 @@ class OWLDumper(Dumper):
                         axiom_type = AnnotationAssertion
                     else:
                         axiom_type = None
+                        logging.info(f"Cannot determine axiom type for {slot.name}, unprocessed={parents}")
+                        unprocessed_parents += parents
+                        #continue
             logging.debug(f'AXIOM TYPE = {axiom_type}')
-            if not axiom_type:
-                logging.debug(f'No axiom type; skipping {k} for {subj}')
-                continue
-            if is_disjunction:
-                # translate the filler list to a single entry that is a disjunction
-                # TODO: allow for different groupings; for now default to 0
-                level = 0
-                eai.add_operands((level, axiom_type.__name__, ObjectUnionOf.__name__), parents, axiom_annotations)
-                #eai.annotations += axiom_annotations
-                parents = []
-            if is_conjunction:
-                # translate the filler list to a single entry that is a conjunction
-                # TODO: allow for different groupings; for now default to 0
-                level = 0
-                eai.add_operands((level, axiom_type.__name__, ObjectIntersectionOf.__name__), parents, axiom_annotations)
-                #eai.annotations += axiom_annotations
-                parents = []
-            for parent in parents:
-                # TODO: make this more generic
-                if axiom_type == SubClassOf:
-                    logging.debug(f'type(subj) = {type(subj)} // {subj}')
-                    if isinstance(subj, AnonymousIndividual):
-                        axiom = None
+            if axiom_type:
+                # special case handling of conjunctions and disjunctions;
+                # these are added to the index, to be processed at the entity level
+                if is_disjunction:
+                    # translate the filler list to a single entry that is a disjunction
+                    # TODO: allow for different groupings; for now default to 0
+                    level = 0
+                    eai.add_operands((level, axiom_type.__name__, ObjectUnionOf.__name__), parents, axiom_annotations)
+                    #eai.annotations += axiom_annotations
+                    parents = []
+                if is_conjunction:
+                    # translate the filler list to a single entry that is a conjunction
+                    # TODO: allow for different groupings; for now default to 0
+                    level = 0
+                    eai.add_operands((level, axiom_type.__name__, ObjectIntersectionOf.__name__), parents, axiom_annotations)
+                    #eai.annotations += axiom_annotations
+                    parents = []
+                for parent in parents:
+                    # Note: when considering making this more generic,
+                    # bear in mind that order or number of arguments may vary
+                    if axiom_type == SubClassOf:
+                        logging.debug(f'type(subj) = {type(subj)} // {subj}')
+                        if isinstance(subj, AnonymousIndividual):
+                            axiom = None
+                        else:
+                            axiom = SubClassOf(subj, parent)
+                    elif axiom_type == SubObjectPropertyOf:
+                        axiom = SubObjectPropertyOf(subj, parent)
+                    elif axiom_type == EquivalentClasses:
+                        axiom = EquivalentClasses(subj, parent)
+                    elif axiom_type == ClassAssertion:
+                        # Note: ClassAssertion axioms are "inverted"
+                        axiom = ClassAssertion(parent, subj)
+                    elif axiom_type == ObjectPropertyAssertion:
+                        axiom = ObjectPropertyAssertion(slot_uri, subj, parent)
+                    elif axiom_type == InverseObjectProperties:
+                        axiom = InverseObjectProperties(subj, parent)
+                    elif axiom_type == ObjectPropertyDomain:
+                        axiom = ObjectPropertyDomain(subj, parent)
+                    elif axiom_type == ObjectPropertyRange:
+                        axiom = ObjectPropertyRange(subj, parent)
+                    elif axiom_type == AnnotationAssertion:
+                        axiom = AnnotationAssertion(slot_uri, subj, parent)
                     else:
-                        axiom = SubClassOf(subj, parent)
-                elif axiom_type == SubObjectPropertyOf:
-                    axiom = SubObjectPropertyOf(subj, parent)
-                elif axiom_type == EquivalentClasses:
-                    axiom = EquivalentClasses(subj, parent)
-                elif axiom_type == ClassAssertion:
-                    axiom = ClassAssertion(parent, subj)
-                elif axiom_type == ObjectPropertyAssertion:
-                    axiom = ObjectPropertyAssertion(slot_uri, subj, parent)
-                elif axiom_type == InverseObjectProperties:
-                    axiom = InverseObjectProperties(subj, parent)
-                elif axiom_type == ObjectPropertyDomain:
-                    axiom = ObjectPropertyDomain(subj, parent)
-                elif axiom_type == ObjectPropertyRange:
-                    axiom = ObjectPropertyRange(subj, parent)
-                elif axiom_type == AnnotationAssertion:
-                    axiom = AnnotationAssertion(slot_uri, subj, parent)
-                else:
-                    raise Exception(f'Unknown: {axiom_type}')
-                if axiom is not None:
-                    self.add_axiom(axiom, o, axiom_annotations)
+                        raise Exception(f'Unknown: {axiom_type}')
+                    if axiom is not None:
+                        self.add_axiom(axiom, o, axiom_annotations)
+                    else:
+                        unprocessed_parents.append(parent)
         # all per-slot axioms have been processed; axioms that span
         # multiple slots are now processed
+        if "IntersectionOf" in cls_interps:
+            if len(unprocessed_parents) == 0:
+                raise ValueError(f"Cannot process IntersectionOf with no parents for {element}")
+            if len(unprocessed_parents) == 1:
+                logging.debug(f"Simplifying IntersectionOf(...) to {unprocessed_parents[0]}")
+                return unprocessed_parents[0]
+            expr = ObjectIntersectionOf(*unprocessed_parents)
+            logging.debug(f"Returning expression {expr} // {eai.operand_list_index.items()}")
+            return expr
         for op_key, operands in eai.operand_list_index.items():
             _, interp, operator = op_key
-            logging.debug(f'EAI {subj}: {interp} => {operator} over {operands}')
-            if operator == ObjectUnionOf.__name__:
+            logging.debug(f'EntityAxiomIndex {subj}: {interp} => {operator} over {operands}')
+            if len(operands) == 0:
+                raise ValueError(f'Too few operands: {operands} for {operator} in {subj}')
+            if len(operands) == 1:
+                logging.debug(f"Simplifying {operator}(...) to {operands[0]}")
+                return operands[0]
+            elif operator == ObjectUnionOf.__name__:
                 expr = ObjectUnionOf(*operands)
             elif operator == ObjectIntersectionOf.__name__:
-                expr = ObjectIntersectionOf(*operands)
+                if len(operands) == 1:
+                    logging.debug(f"Simplifying IntersectionOf(...) to {operands[0]}")
+                    expr = operands[0]
+                else:
+                    expr = ObjectIntersectionOf(*operands)
             else:
                 raise ValueError(f'Cannot handle operator: {operator}')
             if interp == EquivalentClasses.__name__:
                 axiom = EquivalentClasses(subj, expr)
             elif interp == SubClassOf.__name__:
                 axiom = SubClassOf(subj, expr)
+            elif interp == DisjointClasses.__name__:
+                axiom = DisjointClasses(subj, expr)
             else:
                 raise ValueError(f'Not handled: {interp}')
+            logging.debug(f'Adding axiom: {axiom}')
             self.add_axiom(axiom, o, eai.annotation_list_index.get(op_key, []))
+        logging.debug(f"Returning {subj}")
         return subj
 
     def add_axiom(self, axiom: Axiom, ontology: Ontology, axiom_annotations: List[Annotation]) -> None:
+        """
+        Add an axiom to the ontology, appending any annotations to the axiom.
+
+        :param axiom:
+        :param ontology:
+        :param axiom_annotations:
+        :return:
+        """
         axiom.annotations = axiom_annotations
         ontology.axioms.append(axiom)
 
@@ -513,8 +531,9 @@ class OWLDumper(Dumper):
         anc_slots = [slot]
         sv = self.schemaview
         for anc_c in sv.class_ancestors(class_name, reflexive=True):
-            induced_slot = sv.induced_slot(slot.name, anc_c)
-            anc_slots.append(induced_slot)
+            if slot.name in sv.class_slots(anc_c):
+                induced_slot = sv.induced_slot(slot.name, anc_c)
+                anc_slots.append(induced_slot)
         for a in sv.slot_ancestors(slot.name, reflexive=True):
             anc_slots.append(sv.get_slot(a))
         for s in anc_slots:
@@ -623,20 +642,20 @@ class OWLDumper(Dumper):
         for prefix, url in schemaview.namespaces().items():
             prefix_lines.append(f'Prefix( {prefix}: = <{url}> )')
         header = "\n".join(prefix_lines)
-        owl_str = f'{header}\nOntology(\n{owl_str}\n)'
+        owl_str = f'{header}\nOntology(<http://example.org>\n{owl_str}\n)'
         logging.debug(owl_str)
         try:
             doc = to_python(owl_str)
         except Exception as e:
             logging.error(f'Error parsing generated OWL: {owl_str}')
             raise e
-        from funowl.writers.FunctionalWriter import FunctionalWriter
-        from rdflib import Graph
-        g = Graph()
-        for p in doc.prefixDeclarations:
-            g.namespace_manager.bind(p.prefixName, p.fullIRI)
-        fw = FunctionalWriter(g=g)
-        owl_str_roundtrip = doc.to_functional(fw)
+        #from funowl.writers.FunctionalWriter import FunctionalWriter
+        #from rdflib import Graph
+        #g = Graph()
+        #for p in doc.prefixDeclarations:
+        #    g.namespace_manager.bind(p.prefixName, p.fullIRI)
+        #fw = FunctionalWriter(g=g)
+        #owl_str_roundtrip = doc.to_functional(fw)
         #logging.debug(f'ROUNDTRIP = {owl_str_roundtrip}')
         return doc
 
@@ -656,14 +675,31 @@ class OWLDumper(Dumper):
         logging.debug(f'AXIOMS >> = {axioms}')
         self.ontology.axioms += axioms
 
-    def add_axioms_from_template(self, template_ann: Union[str, meta.Annotation], element: YAMLRoot, val: Any = None):
+    def add_axioms_from_template(self, template_ann: Union[str, meta.Annotation], element: YAMLRoot, val: Any = None, schema: SchemaDefinition = None):
         # TODO: simplify, change arg to str
         d = self._element_to_template_dict(element, val)
         if isinstance(template_ann, str):
             tstr = template_ann
         else:
             tstr = template_ann.value
+        def tr(e: YAMLRoot):
+            expr = self.transform(e, schema=schema, is_element_an_object=False)
+            fw = FunctionalWriter()
+            logging.debug(f"template.transform({e}) DIRECT = {expr}")
+            owl_str = str(expr.to_functional(fw))
+            logging.debug(f"template.transform({e}) = {owl_str}")
+            return owl_str
+        if "tr" in d:
+            d["_tr"] = tr
+        else:
+            d["tr"] = tr
         jt = Template(tstr)
+
+        def _tr(x):
+            fw = FunctionalWriter()
+            expr = self.transform(x, schema)
+            return expr.to_functional(fw)
+        d["tr"] = _tr
         owl_str = jt.render(**d)
         axioms = self.parse_axioms_string(owl_str).ontology.axioms
         self.ontology.axioms += axioms
@@ -692,6 +728,8 @@ class OWLDumper(Dumper):
 
 
 @click.command()
+@click.option("-v", "--verbose", count=True)
+@click.option("-q", "--quiet")
 @click.option('-s', '--schema', required=True,
               help="Path to LinkML schema")
 @click.option("--target-class", "-C",
@@ -700,14 +738,19 @@ class OWLDumper(Dumper):
               help="Path to python datamodel module")
 @click.option("--format", "-f",
               help="Input format (will be inferred from file suffix if not specified)")
-@click.option('-o', '--output', required=True,
+@click.option('-o', '--output',
+              type=click.File(mode="w"),
+              default=sys.stdout,
               help="Path to OWL functional syntax output")
-@click.option("--infer-missing/--no-infer-missing",
+@click.option('-O', '--output-type',
+              type=click.Choice(["ofn", "ttl"]),
+              help="Output format")
+@click.option("--autofill/--no-autofill",
               default=False,
               show_default=True,
               help="If True, fill missing data slots using string_serialization")
 @click.argument('inputfile')
-def cli(inputfile: str, schema: str, target_class, module, output, format, infer_missing: bool, **args):
+def cli(inputfile: str, schema: str, target_class, module, output, output_type, format, autofill: bool, verbose: int, quiet: bool, **args):
     """
     Dump LinkML instance data as OWL
 
@@ -715,20 +758,32 @@ def cli(inputfile: str, schema: str, target_class, module, output, format, infer
 
     Convert a CSV to OWL
 
-        linkml-data2owl -s tests/inputs/owl_dumper_test.yaml tests/inputs/parts.csv -o parts.ofn
+        linkml-data2owl -s owl_dumper_test.yaml parts.csv -o parts.ofn
 
-        Note in this example, there must be a class type designator column `@type` in the CSV
+    Note in this example, there must be a class type designator column `@type` in the CSV
 
     Convert a CSV to OWL, homogeneous classes:
 
-        linkml-data2owl -C EquivGenusAndPartOf -s tests/inputs/owl_dumper_test.yaml \
-            tests/inputs/parts_implicit_type.csv -o parts.ofn
+        linkml-data2owl -C EquivGenusAndPartOf -s owl_dumper_test.yaml \
+            parts_implicit_type.csv -o parts.ofn
 
     Convert YAML or JSON to OWL:
 
-        linkml-data2owl -s tests/inputs/owl_dumper_test.yaml tests/inputs/owl_dumper_test_data.yaml -o ont.ofn
+        linkml-data2owl -s owl_dumper_test.yaml owl_dumper_test_data.yaml -o ont.ofn
 
+    More documentation:
+
+        https://linkml.io/linkml-owl/
     """
+    logger = logging.getLogger()
+    if verbose >= 2:
+        logger.setLevel(level=logging.DEBUG)
+    elif verbose == 1:
+        logger.setLevel(level=logging.INFO)
+    else:
+        logger.setLevel(level=logging.WARNING)
+    if quiet:
+        logger.setLevel(level=logging.ERROR)
     if module is None:
         if schema is None:
             raise Exception('must pass one of module OR schema')
@@ -740,11 +795,10 @@ def cli(inputfile: str, schema: str, target_class, module, output, format, infer
     element = load_structured_file(inputfile, target_class=target_class, python_module=python_module, schemaview=sv, fmt=format)
 
     dumper = OWLDumper()
-    if infer_missing:
-        dumper.infer_missing_values = True
-    doc = dumper.dumps(element, schemaview=sv)
-    with open(output, 'w') as stream:
-        stream.write(str(doc))
+    if autofill:
+        dumper.autofill = True
+    doc = dumper.dumps(element, schemaview=sv, output_type=output_type)
+    output.write(str(doc))
 
 
 if __name__ == '__main__':
